var mongoose = require('mongoose')
  , Schema = mongoose.Schema
  , MongooseArray = mongoose.Types.Array
  , merge = require('./utils').merge;

/**
 * Stores the references from its paths
 * into the _relations property.
 *
 * @return {Object}
 * @api private
 */

Schema.prototype._extractReferences = function() {
  var relations = {many: [], one: []};
  for (var path in this.paths) {
    // Verify if the reference is in an array or not
    var isArray = Array.isArray(this.paths[path].options.type)
      , ref;
    
    // Extract the referred model
    if ((isArray && this.paths[path].caster && this.paths[path].caster.options))
      ref = this.paths[path].caster.options.ref;
    else if (!isArray && this.paths[path].options.ref)
      ref = this.paths[path].options.ref;
    else
      ref = null
    
    if (ref) {
      isArray
        ? this.paths[path].many = ref
        : this.paths[path].one = ref
    }
  }
  return this;
};

/**
 * Retrieve first path of schema referencing `ref`
 *
 * @param {String} ref [reference]
 * @return {Object} / {null}
 *   name: String
 *   isArray: Boolean
 * @api private
 */
Schema.prototype._getPathFor = function(ref) {
  var pathName, pathIsArray;
  for (var path in this.paths) {
    // Verify if the reference is in an array or not
    var isArray = Array.isArray(this.paths[path].options.type)
      , pathRef;
    
    // Extract the referred model
    if ((isArray && this.paths[path].caster && this.paths[path].caster.options))
      pathRef = this.paths[path].caster.options.ref;
    else if (!isArray && this.paths[path].options.ref)
      pathRef = this.paths[path].options.ref;
    else
      pathRef = null
    
    if (pathRef && pathRef === ref) {
      pathName = path;
      pathIsArray = isArray;
      break;
    }
  }
  return pathRef
    ? {name: pathName, isArray: pathIsArray}
    : null
};

/**
 * Finds the path referencing supplied model name
 *
 * @param {String} modelName
 * @return {Object}
 *   @param {String} type
 *   @param {String} name
 * @api private
 */
Schema.prototype._findPathReferencing = function(modelName) {
  for (var path in this.paths) {
    var options = this.paths[path].options;
    if (options.belongsTo === modelName || options.habtm === modelName) {
      var type = Array.isArray(options.type)
                  ? "habtm"
                  : "belongsTo"
      return {type: type, name: path};
      break;
    }
  }
};

/**
 * Create a child document and add it to the parent `Array`
 *
 * @param {Object|Array} objs [object(s) to create]
 * @param {Functions} callback [passed: (err, parent, created children)]
 * @api public
 */
MongooseArray.prototype.create = function(objs, callback){
  if (!(this._schema) && !(this._schema.options.hasMany || this._schema.options.habtm || this._schema.options.hasOne))
    return callback(new Error("Path doesn't contain a reference"));
  
  var self = this
    , parent = this._parent
    , childModelName = this._schema.options.hasMany || this._schema.options.habtm
    , childModel = parent.model(childModelName)
    , childSchema = childModel.schema
    , parentModelName = parent.constructor.modelName
    , childPath = childSchema._findPathReferencing(parentModelName);
  
  // You *need* a reference in the child `Document`
  if (!childPath)
    throw new Error("Parent model not referenced anywhere in the Child Schema");
  
  /**
   * DRY create function for the child `Document`
   *
   * @param {Object} obj [properties of the `Document`]
   * @param {Function} callback
   * @api very private
   */
  var create = function(obj, cb) {
    obj = new childModel(obj);
    // HABTM or belongsTo?
    if (childPath.type === "habtm")
      obj[childPath.name].push(parent);
    else
      obj[childPath.name] = parent;
    obj.save(cb)
  }
  
  // If we're provided an `Array`, we need to iterate
  if (Array.isArray(objs)) {
    var created = []
      , total = objs.length;
    objs.forEach(function(obj, i){
      create(obj, function(err, obj){
        if (err) {
          // Empty the array and return the error,
          // effectively breaking the loop.
          objs.splice(i, objs.length - i);
          return callback(err);
        }
        // Store the created records;
        created.push(obj);
        parent[self._path].push(obj);
        --total || parent.save(function(err, parent){
          if (err) return callback(err);
          return callback(null, parent, created);
        });
      });
    });
  } else {
    // Only one object needs creation.
    // Going for it then!
    create(objs, function(err, obj){
      if (err) return callback(err);
      parent[self._path].push(obj);
      parent.save(function(err, parent){
        if (err) return callback(err);
        return callback(null, parent, obj);
      });
    });
  }
};

/**
 * Find children documents
 *
 * *This is a copy of Model.find w/ added error throwing and such*
 */
MongooseArray.prototype.find = function(conditions, fields, options, callback){
  if (!(this._schema) && !(this._schema.options.hasMany || this._schema.options.habtm || this._schema.options.hasOne))
    return callback(new Error("Path doesn't contain a reference"));

  // Copied from `Model.find`
  if ('function' == typeof conditions) {
    callback = conditions;
    conditions = {};
    fields = null;
    options = null;
  } else if ('function' == typeof fields) {
    callback = fields;
    fields = null;
    options = null;
  } else if ('function' == typeof options) {
    callback = options;
    options = null;
  } else {
    conditions = {};
  }
  
  var parent = this._parent
    , childModel = parent.model(this._schema.options.hasMany || this._schema.options.habtm)
    , childPath = childModel.schema._findPathReferencing(parent.constructor.modelName);
  
  // You *need* a reference in the child `Document`
  if (!childPath)
    throw new Error("Parent model not referenced anywhere in the Child Schema");
  
  var safeConditions = {}
  safeConditions[childPath.name] = parent._id;
  
  merge(safeConditions, conditions)
  merge(safeConditions, {_id: {$in: parent[this._path]}})
  
  var query = new mongoose.Query(safeConditions, options).select(fields).bind(childModel, 'find');

  if ('undefined' === typeof callback)
    return query;

  return query.find(callback);
};

/**
 * Adds the relationships to the Schema
 *
 * @param {Object} relation
 * @return {Schema}
 * @api private
 */
Schema.prototype._addRelationships = function(relationships) {
  var self = this;
  var addRelationship = function(rel) {
    var type
      , through = rel.through;
    if (!through)
      throw new Error("You need to specify through which path is this relationship referenced.");
    
    for (var key in rel) {
      if (key !== "through") {
        type = key;
        break;
      }
    }
    
    if (!type)
      throw new Error("You need to specify a type of relationship.");
    if (type !== "hasMany" && type !== "belongsTo" && type !== "hasOne" && type !== "habtm")
      throw new Error("Relationship type incorrect.");
    
    // Add it to the path's options
    self.paths[rel.through].options[type] = rel[type];
  }
  
  if (Array.isArray(relationships)) {
    relationships.forEach(function(rel){
      addRelationship(rel);
    });
  } else {
    addRelationship(relationships);
  }
  
  return this;
}

module.exports = function plugin(schema, relationships){
  if (!relationships) throw new Error("Please specify your Schema's relationships.");
  
  schema._addRelationships(relationships);
  
  return schema;
};

