var start = require('./common')
  , should = require('should')
  , mongoose = start.mongoose
  , random = require('../lib/utils').random
  , Schema = mongoose.Schema
  , ObjectId = Schema.ObjectId
  , relationships = require('../index');

/**
 * Post Schema
 * "has many categories & belongs to one author"
 */
var PostSchema = new Schema({
    title      : String
  , body       : Number
  , author     : ObjectId
  , categories : [ObjectId]
});

/**
 * Category Schema
 * "has many posts"
 */

var CategorySchema = new Schema({
    name: String
  , posts : [ObjectId]
});

/**
 * User Schema
 * "has many posts"
 */
var UserSchema = new Schema({
    name  : String
  , posts : [ObjectId]
  , tags  : [String]
});

/**
 * Attach the plugin to the schemas
 */
PostSchema.plugin(relationships, [
    {
        belongsTo : "User"
      , through   : "author"
    },
    {
        habtm   : "Category"
      , through : "categories"
    }
]);
CategorySchema.plugin(relationships, {
    habtm   : "Post"
  , through : "posts"
});
UserSchema.plugin(relationships, {
    hasMany   : "Post"
  , through   : "posts"
});

/**
 * Register the models with Mongoose
 */
mongoose.model('Post', PostSchema)
mongoose.model('Category', CategorySchema)
mongoose.model('User', UserSchema);

var posts = 'posts_' + random()
  , users = 'users_' + random()
  , categories = 'categories_' + random();

/**
 * Tests
 */

module.exports = {

  'test parent schema habtm path': function() {
    CategorySchema.paths['posts'].options.habtm.should.equal('Post');
  },

  'test child schema habtm path': function() {
    PostSchema.paths['categories'].options.habtm.should.equal('Category');
  },

  'test presence of create method in both document': function() {
    var db = start()
      , Category = db.model('Category', categories)
      , category = new Category()
      , Post = db.model('Post', posts)
      , post = new Post();

    category.posts.should.respondTo('create');
    post.categories.should.respondTo('create');
    
    db.close()
  },

  'test create one child document': function() {
    var db = start()
      , Category = db.model('Category', categories)
      , category = new Category()
      , Post = db.model('Post', posts);

    var post = {title: "Easy relationships with mongoose-relationships"};

    category.posts.create(post, function(err, category, post){
      should.strictEqual(err, null);

      category.should.be.an.instanceof(Category);
      category.posts.should.have.length(1);

      category.posts[0].should.equal(post._id);

      post.should.be.an.instanceof(Post);
      post.title.should.equal("Easy relationships with mongoose-relationships")
      post.categories.should.contain(category._id);

      db.close();
    });
  },

  'test create many child documents': function(){
    var db = start()
      , Category = db.model('Category', categories)
      , category = new Category()
      , Post = db.model('Post', posts);

    var posts = [
        {title: "Blog post #1"}
      , {title: "Blog post #2"}
    ]

    category.posts.create(posts, function(err, category, posts){
      should.strictEqual(err, null);

      category.posts.should.have.length(2);

      posts.should.have.length(2);
      posts.forEach(function(post){
        category.posts.should.contain(post._id)
        post.should.be.an.instanceof(Post);
        post.categories.should.contain(category._id);
      });

      db.close();
    });
  },

  'test find children documents': function(){
    var db = start()
      , Category = db.model('Category', categories)
      , category = new Category()
      , Post = db.model('Post', posts);

    var posts = [
        {title: "Blog post #1"}
      , {title: "Blog post #2"}
    ]

    category.posts.create(posts, function(err, category, posts){
      var find = category.posts.find({}, function(err, newPosts){
        should.strictEqual(err, null);

        find.should.be.an.instanceof(mongoose.Query);
        find._conditions.should.have.property('_id');
        find._conditions.should.have.property('categories');
        find._conditions._id['$in'].should.be.an.instanceof(Array);

        newPosts.should.have.length(2);
        newPosts.forEach(function(post){
          category.posts.should.contain(post._id)
          post.should.be.an.instanceof(Post);
          post.categories.should.contain(category._id);
        });

        find.find({title: "Blog post #1"}, function(err, otherPosts){

          find._conditions.title.should.equal("Blog post #1");
          find._conditions.should.have.property('_id');

          otherPosts.should.have.length(1);

          var post = otherPosts[0];
          post.title.should.equal("Blog post #1");

          db.close();

        });
      });
    });
  }
};