var start = require('./common')
  , should = require('should')
  , mongoose = start.mongoose
  , random = require('../lib/utils').random
  , Schema = mongoose.Schema
  , ObjectId = Schema.ObjectId
  , relationships = require('../index');

/**
 * Blog Post Schema
 * "has many categories & belongs to one author"
 */
var PostSchema = new Schema({
    title      : String
  , body       : Number
  , author     : ObjectId
});

/**
 * User Schema
 * "has many posts"
 */
var UserSchema = new Schema({
    name  : String
  , posts : [ObjectId]
  , tags  : [String]
});

/**
 * Attach the plugin to the schemas
 */
PostSchema.plugin(relationships, {
    belongsTo : "User"
  , through   : "author"
});
UserSchema.plugin(relationships, {
    hasMany   : "Post"
  , through   : "posts"
});

/**
 * Register the models with Mongoose
 */
mongoose.model('Post', PostSchema);
mongoose.model('User', UserSchema);

var posts = 'posts_' + random()
  , users = 'users_' + random();

/**
 * Tests
 */

module.exports = {
  
  'test parent schema hasMany path': function() {
    UserSchema.paths['posts'].options.hasMany.should.equal('Post');
  },
  
  'test child schema belongsTo path': function() {
    PostSchema.paths['author'].options.belongsTo.should.equal('User');
  },
  
  'test presence of create method in parent document': function() {
    var db = start()
      , User = db.model('User', users)
      , user = new User();

    user.posts.should.respondTo('create');
    db.close()
  },
  
  'test create one child document': function() {
    var db = start()
      , User = db.model('User', users)
      , Post = db.model('Post', posts)
      , user = new User();
    
    var post = {title: "Easy relationships with mongoose-relationships"};
    
    user.posts.create(post, function(err, user, post){
      should.strictEqual(err, null);
      
      user.should.be.an.instanceof(User);
      user.posts.should.have.length(1);

      user.posts[0].should.equal(post._id);
      
      post.should.be.an.instanceof(Post);
      post.title.should.equal("Easy relationships with mongoose-relationships")
      post.author.should.equal(user._id);
      
      db.close();
    });
  },
  
  'test create many child documents': function(){
    var db = start()
      , User = db.model('User', users)
      , Post = db.model('Post', posts)
      , user = new User();
    
    var posts = [
        {title: "Blog post #1"}
      , {title: "Blog post #2"}
    ]
    
    user.posts.create(posts, function(err, user, posts){
      should.strictEqual(err, null);
      
      user.posts.should.have.length(2);
      
      posts.should.have.length(2);
      posts.forEach(function(post){
        user.posts.should.contain(post._id)
        post.should.be.an.instanceof(Post);
        post.author.should.equal(user._id);
      });
      
      db.close();
    });
  },
  
  'test find children documents': function(){
    var db = start()
      , User = db.model('User', users)
      , Post = db.model('Post', posts)
      , user = new User();
    
    var posts = [
        {title: "Blog post #1"}
      , {title: "Blog post #2"}
    ]
    
    user.posts.create(posts, function(err, user, posts){
      var find = user.posts.find({}, function(err, newPosts){
        should.strictEqual(err, null);
        
        find.should.be.an.instanceof(mongoose.Query);
        find._conditions.should.have.property('_id');
        find._conditions.should.have.property('author');
        find._conditions._id['$in'].should.be.an.instanceof(Array);
        
        newPosts.should.have.length(2);
        newPosts.forEach(function(post){
          user.posts.should.contain(post._id)
          post.should.be.an.instanceof(Post);
          post.author.should.eql(user._id);
        });
        
        find.find({title: "Blog post #1"}, function(err, otherPosts){
          
          find._conditions.title.should.equal("Blog post #1");
          find._conditions.should.have.property('_id');
          
          otherPosts.should.have.length(1);
          
          var post = otherPosts[0];
          post.title.should.equal("Blog post #1");
          
          db.close();
          
        });
      });
    });
  }
};